#include "composite.h"
#include <vector>
#include <algorithm>
#include "simulation.h"
#include "event.h"

using namespace std;

extern FILE* logFile;
extern bool logging;

Composite::Composite( Simulation* sim0,double mass,bool ignoreGravity0 ) : Object(sim0), m( mass ), ignoreGravity(ignoreGravity0) {
}
;

Composite::Composite( Composite &x ): Object(sim) {
	objs.swap( x.objs );
};
Composite::~Composite() {
	for ( int i = 0;i < objs.size();i++ )
		delete objs[ i ];
};

double Composite::mass() const {
	return m;
};

inline Composite::CollisionInfo Composite::findFirstCollision( const Particle& p ) const {
	Composite::CollisionInfo r = {infinity, 0};
	for ( int i = 0;i < objs.size();i++ ) {
		objs[ i ] ->vx = vx;
		objs[ i ] ->vy = vy;
		double t1 = objs[ i ] ->computeCollisionTime( p );
		if ( t1 < r.time ) {
			r.time = t1;
			r.obj = objs[ i ];
		}
	}
	return r;
}

double Composite::computeCollisionTime( const Particle& p ) const {
	Composite::CollisionInfo info = findFirstCollision( p );
	if ( info.obj ) {
		if (logging)
			fprintf( logFile, "id:%2d, part %2d could collide with %2d\n", getId(), info.obj->getId(), p.getId() );
	};
	return info.time;
};

void Composite::collisionDirection( const Particle& p, double &dx, double &dy ) {
	CollisionInfo ci = findFirstCollision( p );
	ci.obj->collisionDirection( p, dx, dy );
}

// Composite is a passive object, it does not look for collisions
// But it can change its velocity; it has to inform nearby particles about that
void Composite::findCollision(bool delCollision ) {
	vector<CellId> cells;
	cells.reserve( 9 );
	for ( int i = 0;i < objs.size();i++ ) {
		objs[ i ] ->getOwners(&cells );
	};
	typedef vector<CellId>::iterator iter;
	sort( cells.begin(), cells.end() );
	iter last = unique( cells.begin(), cells.end() );
	//	iter last=cells.end();
	for ( iter ci = cells.begin();ci != last;ci++ ) {
		Cell &cell = sim->getCell( *ci );
		sim->synchronize( *ci );
		for ( List<Particle>::iterator it = cell.getParticleIter();
		        it.valid();++it ) {
			it->considerCollisionWith(*this );

		};
	};
	Rectangle r=boundingRect();
	
};

void Composite::writeLog( FILE* file ) {
	fprintf( file, "id:%2d, vx: %lf, vy: %lf\n", getId(), vx, vy );
	for ( int i = 0;i < objs.size();i++ ) {
		fprintf( file, "\t" );
		objs[ i ] ->writeLog( file );
	};

};
bool Composite::testCollision( Particle const& p ) const {
	for ( int i = 0;i < objs.size();i++ )
		if ( objs[ i ] ->testCollision( p ) )
			return true;
	return false;
};

void Composite::addObject( Object* o ) {
	objs.push_back( o );
};

void Composite::draw( Canvas& c ) const {
	for ( int i = 0;i < objs.size();i++ )
		objs[ i ] ->draw( c );
};

void Composite::run( double dt ) {
	for ( int i = 0;i < objs.size();i++ ) {
		objs[ i ] ->vx = vx;
		objs[ i ] ->vy = vy;
		objs[ i ] ->run( dt );
	};
	if (!ignoreGravity) {
		double g=sim->getGravity();
		vy-=g*dt;
	};
};

Rectangle Composite::boundingRect() const {
	if (objs.size()==0) 
		return Rectangle(0,0,0,0);
	Rectangle r=objs[0]->boundingRect() ;
	for ( int i=1;i<objs.size();i++) {
		Rectangle s=objs[i]->boundingRect();
		r.x1=min(r.x1,s.x1);
		r.y1=min(r.y1,s.y1);
		
		r.x2=max(r.x2,s.x2);
		r.y2=max(r.y2,s.y2);
	};
	return r;
};

void Composite::getOwners( std::vector<CellId> *v ) const {
	for ( int i = 0;i < objs.size();i++ ) {
		objs[ i ] ->getOwners( v );
	};
};

void Composite::doCollisionPrivate(Particle &p)
{
	CollisionInfo ci = findFirstCollision( p );
	if (!ci.obj) return;
	ci.obj->doCollisionWith(p);
	vx=ci.obj->vx;
	vy=ci.obj->vy;
	for ( int i = 0;i < objs.size();i++ ) {
		objs[i]->vx=vx;
		objs[i]->vy=vy;
	};
		
}
