#ifndef CQUEUE_H
#define CQUEUE_H
//#define _GLIBCXX_DEBUG
#include "common.h"
#include <vector>
#include <cmath>
#include <cassert>
#include <cstdio>

#include "queueitem.h"



/*	KeyExtractor should contain:
	typename KeyExtractor::Key 
	KeyExtractor::Key KeyExtractor::getKey(T* const)  
*/


template<class T,class KeyExtractor>
class SortedArray {
	std::vector<T*> data;
	typedef typename KeyExtractor::Key Key;
public:
	//invariant: items in data are sorted in descending order
	void checkInvariants(KeyExtractor e) {
		for (size_t i=1;i<data.size();i++)
			assert(e.getKey(data[i-1])>=e.getKey(data[i]));
	};
	T* min(KeyExtractor e) const {return data.back();}
	Key minkey(KeyExtractor e) const {return empty()? infinity : e.getKey(min(e));}
	void pop(KeyExtractor e) {data.pop_back();}
	void push(KeyExtractor e,T* const x) {
		data.push_back(x);
		for (size_t i=data.size()-1;i>0 && e.getKey(data[i-1])<e.getKey(data[i]);i--)
			swap(data[i-1],data[i]);
	};
	bool empty() const {return data.empty();}
	void remove(KeyExtractor,T* const x) {
		size_t i;
		for (i=0;i<data.size() && data[i]!=x;i++) ;
		if (i==data.size()) return;  //if x was not found
		// now data[i]==x
		for (size_t j=i;j<data.size()-1;j++) data[j]=data[j+1];
		data.pop_back();
	}
	size_t size() const {return data.size();}
	void reserve(size_t n) {data.reserve(n);}
};

template<class T,class KeyExtractor>
class CQueue {
	typedef typename KeyExtractor::Key Key;
//	typedef Heap<T,Traits> Bucket;
    KeyExtractor ke;
	inline Key getKey(T* const x) const {return ke.getKey(x);}
	typedef std::vector<SortedArray<T,KeyExtractor> > BucketArray;
	BucketArray bucket;
	Key dlen;
	mutable size_t day,year;
	size_t count,minBuckets,maxBuckets;
	const double itemsPerBucket;

	size_t yearOf(const Key &x) const {
		return size_t(x/(dlen*bucket.size()));
	};
	size_t dayOf(const Key &x) const {
		Key ylen=dlen*bucket.size();
		Key r=fmod(x,ylen);
		while (r<0) r+=ylen;
		return size_t(r/dlen);
	}
	void find_min() const { // advance day,year so that bucket[day] contains the minimum
		assert(!empty());
		size_t minday=day;
		size_t day0=day;
		size_t buckets=bucket.size();
		Key ystart=year*dlen*buckets;
		Key yend=(year+1)*dlen*buckets;
		do {
			if (!bucket[day].empty()) {
				Key x=bucket[day].minkey(ke);
				if (ystart<=x && x<yend) 
					return;  //first event in current year 
				if (x<bucket[minday].minkey(ke)) minday=day;
			}
			if (++day==buckets) {
				day=0; ++year;
				ystart=yend;
				yend=(year+1)*dlen*buckets;
			}
		} while (day!=day0);
		assert(!bucket[minday].empty());  //if the cqueue is not empty, there must be minimal element
		year=yearOf(bucket[minday].minkey(ke));
		day=minday;
	}
	
	Key newDayLen() {
		if (count<2) return 1.0;
		int nsamples;
		if (count<=5) nsamples=count;
		else nsamples=5+count/10;
		if (nsamples>25) nsamples=25;
		T* samples[25];	
		for (int i=0;i<nsamples;i++) {
			samples[i]=getMin();
			popNoResize();
		};		
		Key sep1=getKey(samples[nsamples-1])-getKey(samples[0]);
		sep1/=nsamples-1;
		
		int nsamples2=0;
		Key sep2=0;
		for (int i=0;i<nsamples-1;i++) {
			Key delta=getKey(samples[i+1])-getKey(samples[i]);
			if (delta<2*sep1) {sep2+=delta; nsamples2++;}
		}
		sep2/=nsamples2;
		
		for (int i=nsamples-1;i>=0;i--) 
			pushNoResize(samples[i]);
		return sep2;
	};

	void popNoResize() {
		find_min();
		bucket[day].min(ke)->QueueItem::position=-1;
//		fprintf(stderr,"popMin: %lf\n",getKey(bucket[day].min(ke)));
		checkInvariants();
		bucket[day].pop(ke);
		count--;
	}
	void pushNoResize(T* const x) {
		size_t d=dayOf(getKey(x));
		size_t y=yearOf(getKey(x));
		x->QueueItem::position=d;
		bucket[d].push(ke,x);
		if (y<year || (y==year && d<day)) {  //if x is before current date
			year=y;
			day=d;
		};
		count++;
	};

	void checkTooSmall() {
		if (count<minBuckets*itemsPerBucket && minBuckets>1) {
			minBuckets/=2; maxBuckets/=2;
			if (minBuckets<1) minBuckets=1;
			resize((minBuckets+maxBuckets)/2);
		}
	};
	void checkTooLarge() {
		if (count>2*minBuckets*itemsPerBucket) {
			minBuckets*=2; maxBuckets*=2;
			resize((minBuckets+maxBuckets)/2);
		};
	}

public:	
	void resize(size_t newsize) {
		dlen=newDayLen();
		BucketArray oldbucket;
		oldbucket.swap(bucket); count=0;
		
		bucket.resize(newsize);
        int bs=itemsPerBucket;
        for (int i=0;i<bucket.size();i++)
            bucket[i].reserve(bs);
		for (int i=0;i<oldbucket.size();i++) {
			while (!oldbucket[i].empty()) {
				pushNoResize(oldbucket[i].min(ke));
				oldbucket[i].pop(ke);
			}
            
		};
        
//		checkInvariants();
	};
	
	CQueue(KeyExtractor ke0,double ipb):bucket(4),dlen(1),count(0),day(0),year(0),itemsPerBucket(ipb),minBuckets(1),
		maxBuckets(4),ke(ke0) {}
	T* getMin() const {
		find_min();
        T* const result=bucket[day].min(ke);
		return result;
	}
	void popMin() {
		popNoResize();
		checkTooSmall();
	}
	void insert(T* const x) {
		pushNoResize(x);
		checkTooLarge();
	};
	bool empty() const {return count==0;}
	void remove(T* x) {
		if (x->QueueItem::position==-1) return;
		bucket[x->QueueItem::position].remove(ke,x);
		x->QueueItem::position=-1;
		count--;
		checkTooSmall();
	};
	inline void reset(T* x) {
		if (x->QueueItem::position==-1) return;
		remove(x);
		insert(x);
	};
	
	void checkInvariants() {
		size_t sum=0;
		for (size_t i=0;i<bucket.size();i++) {
			sum+=bucket[i].size();
			bucket[i].checkInvariants(ke);
		};
		assert(sum==count);
	};
};

//template<class T,class Traits> double CQueue<T,Traits>::itemsPerBucket;
#endif
