
//#include "all.h"
#define QT_THREAD_SUPPORT
#include "simulation.h" // class's header file
#include "heap.h" 
//#include "cqueue.h"
#include "myassert.h"
#include <cstdio>
#include "common.h"
#include <cmath>
#include "safevector.h"
#include "timer.h"
#include <stdexcept>
#include <qmutex.h>
#include "event.h"
#include "composite.h"
#include <boost/intrusive_ptr.hpp>

#include "refCounter.cpp"

using boost::intrusive_ptr;

//for debugging purposes
bool logging = false;
int assert_level = 1;
double stop_time = 0;
FILE *logFile;


using namespace std;

/*
 
struct ParticleTraits {
	typedef double Key;
	static inline Key getKey(Particle* const p) {return p->getCollisionTime();}
	};
 
struct ParticleKeyExtractor {
    ParticleKeyExtractor(Simulation* sim0):sim(sim0) {}
    typedef double Key;
    inline Key getKey(Particle* const p) const {
        return p->getCollisionTime()+sim->getCell(p->owner).getClock();
	}
	private:
    Simulation* sim;
	};
*/
struct SimulationData {
private:
	//this is private to ensure syncronization of composites
	double totalTime;
public:
	double gravity;
	vector<Particle*> particles;
	vector<Wall*> walls;
	
	// composites are always synchronized with totalTime
	vector<Composite*> composites;
	vector<Cell> cells;
	CellId emptyCell;
	vector<CellBorder> verticalBorders, horizBorders;
	double minx, miny, maxx, maxy, cellWidth, cellHeight;
	int cellCount, sqrtCellCount;
	Heap<Event, Compare<EventKeyExtractor, Event>, intrusive_ptr<Event> > events;
	//	CQueue<Event,EventKeyExtractor> events;
	int collisionCount;
	QMutex running, should_stop_mutex;
	volatile bool should_stop;
	SimulationData( int particleCount, int wallCount, Simulation* sim ) :
			totalTime( 0 )
			, events( particleCount, Compare<EventKeyExtractor, Event>( EventKeyExtractor() ) )
			//		,events(EventKeyExtractor(),4)
	{};
	double getTotalTime() {
		return totalTime;
	};
	void setTotalTime( double newTime ) {
		double dt = newTime - totalTime;
		assert( dt > -eps );
		for ( int i = 0;i < composites.size();i++ )
			composites[ i ] ->run( dt );
		totalTime = newTime;
	};
};

Simulation::Simulation( int particleCount, int wallCount, int cellCount, double maxx, double maxy, double gravity ):data(new SimulationData( particleCount, wallCount, this )) {
	data->gravity = gravity;
	data->particles.reserve( particleCount );
	data->walls.reserve( wallCount );
	data->composites.reserve(10);
	data->sqrtCellCount = ( int ) ceil( sqrt( ( double ) cellCount ) );
	data->cellCount = ( int ) sqr( data->sqrtCellCount );
	data->cells.reserve( data->cellCount + 1 );
	for ( int i = 0;i != data->cellCount + 1;i++ )
		data->cells.push_back( Cell() );
	data->minx = data->miny = 0;
	data->maxx = maxx;
	data->maxy = maxy;
	data->cellWidth = ( data->maxx - data->minx ) / data->sqrtCellCount;
	data->cellHeight = ( data->maxy - data->miny ) / data->sqrtCellCount;
	data->emptyCell = data->cellCount;

	// borders of the area
	// therefore they ignore gravity (the last param is true)
	addObject(new Wall( this,wVertical, 0, -infinity, infinity, infinity, true )) ;
	addObject(new Wall( this,wVertical, maxx, -infinity, infinity, infinity, true ));
	addObject(new Wall( this,wHorizontal, 0, -infinity, infinity, infinity, true ));
	addObject(new Wall( this,wHorizontal, maxy, -infinity, infinity, infinity, true ));

	initialized = false;
	logFile = fopen( "gas.log", "w" );
	data->should_stop = false;
}

Simulation::~Simulation() {
	stop();
	data->running.lock();
	for ( int i = 0;i < data->particles.size();i++ )
		delete data->particles[ i ];
	for ( int i = 0;i < data->walls.size();i++ )
		delete data->walls[ i ];
	for ( int i = 0;i < data->composites.size();i++ )
		delete data->composites[ i ];
	data->running.unlock();
	fclose( logFile ); 
#ifndef NDEBUG
	// we mark the memory that is going to be free
	memset( this, 0xFF, sizeof( *this ) );
#endif

}

ParticleInfo Simulation::getParticle( int index ) {
	return ParticleInfo( *data->particles[ index ] );
}
Wall& Simulation::getWall( int index ) {
	return * data->walls[ index ];
}

Object& Simulation::getComposite( int index ) {
	return * data->composites[ index ];
}

double Simulation::getMaxX() {
	return data->maxx;
}
double Simulation::getMaxY() {
	return data->maxy;
}

void Simulation::addObject( Particle *p ) {
	data->particles.push_back( p );
	//    data->particles.back().updateOwner(*this);
}

void Simulation::addObject( Wall *w ) {
	data->walls.push_back( w );
}

void Simulation::addObject( Composite *w ) {
	data->composites.push_back( w );
}

int Simulation::getParticleCount() {
	return data->particles.size();
};
int Simulation::getWallCount() {
	return data->walls.size();
};


int Simulation::getCompositeCount() {
	return data->composites.size();
};

/*
void Simulation::reset(Particle &p) {
    data->collisions.reset(&p);
	};    
*/
void Simulation::addEvent( intrusive_ptr<Event> e ) {
	data->events.insert( e );
};
void Simulation::removeEvent( intrusive_ptr<Event> e ) {
	data->events.remove( e );
};


void Simulation::initialize() {
	data->horizBorders.reserve( data->sqrtCellCount );
	data->verticalBorders.reserve( data->sqrtCellCount );
	for ( int i = 0;i != data->sqrtCellCount;i++ ) {
		data->horizBorders.push_back(
		    CellBorder( wHorizontal, ( i + 0.5 ) * data->cellHeight, this ) );
		data->verticalBorders.push_back(
		    CellBorder( wVertical, ( i + 0.5 ) * data->cellWidth, this ) );
	};
	
	Object* o=&data->horizBorders[0];
	o->computeCollisionTime(*data->particles[0]);
	
	
	double maxRadius = 0;
	for ( int i = 0;i != data->particles.size();i++ ) {
		double r = data->particles[ i ] ->radius();
		if ( r > maxRadius )
			maxRadius = r;
	};
	if ( maxRadius > data->cellWidth )
		data->cellWidth = maxRadius;
	if ( maxRadius > data->cellHeight )
		data->cellHeight = maxRadius;
	//now cellWidth, cellHeight are OK, we can insert particles into cells
	for ( int i = 0;i != data->particles.size();i++ )
		data->particles[ i ] ->updateOwner();

	//check for overlapping particles

	if ( testOverlapping() )
		throw std::runtime_error( "Overlapping particles" );

	for ( int i = 0;i != data->particles.size();i++ ) {
		Particle *p = data->particles[ i ];
		p->findCollision( false );
	};
	initialized = true;
};

void Simulation::invalidateCollisionersOf( Object& p, CellId oldOwner ) {
	if ( !valid( oldOwner ) )
		return ;
	for ( int i = 0;i != getNeighbourCount( oldOwner );i++ ) {
		Cell &cell = getCell( getNeighbour( oldOwner, i ) );
		for ( List<Particle>::iterator it = cell.getParticleIter();
		        it.valid();++it ) {
			it->invalidateCollisioner( p );
		};
	};
};

void Simulation::writeLog() {
	if ( logging ) {
		//            data->collisions.test();

		Event & e = *data->events.getMin();
		fprintf( logFile, "Time: %lf\n", data->getTotalTime() );
		fprintf( logFile, "Next collision: %d, time %lf\n",
		         ( ( CollisionEvent& ) e ).getParticle() ->getId(), e.getTime() );
		fprintf( logFile, "Particles:\n" );
		for ( int i = 0;i != data->particles.size();i++ ) {
			data->particles[ i ] ->writeLog( logFile );
			//                data->particles[i].test(*this);
		};
		for ( int i = 0;i != data->composites.size();i++ ) {
			data->composites[ i ] ->writeLog( logFile );
		};
		//        fprintf(logFile,"Walls:\n");
		//        for (int i=0;i!=data->walls.size();i++)
		//            data->walls[i].writeLog(logFile);
		fprintf( logFile, "\n" );
	}
};

double Simulation::getNextCollisionTime() {
	intrusive_ptr<Event> e= data->events.getMin();
	double t1 = e->getTime();
	return t1;
}

void Simulation::run1() {
	intrusive_ptr<Event> e = data->events.getMin();
	data->setTotalTime( e->getTime() );
#	ifndef NDEBUG

	int x;
	if ( data->getTotalTime() > stop_time )
		x = 1;
#	endif

	e->doEvent( this );
	data->events.remove( e );
}

void Simulation::advanceBy( double dt ) {
	QMutexLocker lock ( &data->running )
		;
	if ( !initialized )
		initialize();
	writeLog();
	fflush( logFile );

	data->collisionCount = 0;
	double targetTime = data->getTotalTime() + dt;
	if ( assert_level >= 2 ) {
		test();
		assert( !testOverlapping() );
	};

	while ( true ) {
		data->collisionCount++;
		if ( getNextCollisionTime() > targetTime )
			break;
		run1();
		writeLog();
		fflush( logFile );
	};
	data->setTotalTime( targetTime );
	for ( int i = 0;i != data->cells.size();i++ )
		data->cells[ i ].advanceTo( data->getTotalTime());
}

void Simulation::run() {
	{
		double advance = data->getTotalTime();
		int count = 0;

		data->should_stop_mutex.lock();
		QMutexLocker lock ( &data->running )
			;
		data->should_stop = false;
		data->should_stop_mutex.unlock();
		if ( !initialized )
			initialize();
		if ( assert_level >= 2 )
			test();
		do {
			run1();
			count++;
			//			puts(data->should_stop?"should_stop=yes":"should_stop=no");
		} while ( !data->should_stop );
		advance = data->getTotalTime() - advance;
		//		qDebug("advanced by %lf, %d elementary steps",advance,count);
		for ( int i = 0;i != data->cells.size();i++ )
			data->cells[ i ].advanceTo( data->getTotalTime());
	}
}

void Simulation::stop() {
	data->should_stop = true;
	//	QMutexLocker lock(&data->running);
};

double Simulation::getTotalTime() {
	return data->getTotalTime();
};

CellId Simulation::getNeighbour( CellId ci, int index ) const {
	static const int dx[] = {
	                            -1, 0, 1, -1, 0, 1, -1, 0, 1
	                        };
	static const int dy[] = {
	                            -1, -1, -1, 0, 0, 0, 1, 1, 1
	                        };
	int x = ci % data->sqrtCellCount + dx[ index ];
	int y = ci / data->sqrtCellCount + dy[ index ];
	if ( x >= 0 && x < data->sqrtCellCount && y >= 0 && y < data->sqrtCellCount )
		return x + y * data->sqrtCellCount;
	else
		return data->emptyCell;
};

Cell& Simulation::getCell( CellId ci ) {
	assert( ci >= 0 && ci < data->cells.size() );
	return data->cells[ ci ];
};

CellBorder* Simulation::getBorder( CellId ci, int index ) {
	int x = ci % ( data->sqrtCellCount );
	int y = ci / ( data->sqrtCellCount );
	switch ( index ) {
	case 0:    //lavy okraj
		if ( x > 0 )
			return & data->verticalBorders[ x - 1 ];
		else
			return 0;
	case 1:    //pravy
		if ( x < data->sqrtCellCount - 1 )
			return & data->verticalBorders[ x + 1 ];
		else
			return 0;
	case 2:    //horny
		if ( y > 0 )
			return & data->horizBorders[ y - 1 ];
		else
			return 0;
	case 3:    //dolny
		if ( y < data->sqrtCellCount - 1 )
			return & data->horizBorders[ y + 1 ];
		else
			return 0;
	};
};

CellId Simulation::findOwnerOf( double x, double y ) const {
	if (x<data->minx || y<data->miny || x>=data->maxx || y>=data->maxy)
		return data->cells.size()-1;
	int cx = ( int ) floor( ( x - data->minx ) / data->cellWidth );
	int cy = ( int ) floor( ( y - data->miny ) / data->cellHeight );
	CellId r=cx + cy*data->sqrtCellCount;
	return r;
};

CellId Simulation::findOwnerOf( Particle const &p ) const {
	ParticleInfo pi = ParticleInfo( p );
	return findOwnerOf( pi.x(), pi.y() );
}

int Simulation::getCollisionCount() {
	return data->collisionCount;
};

bool Simulation::valid( CellId ci ) {
	return ci >= 0 && ci < data->cells.size();
};

bool Simulation::test() {
	for ( int i = 0;i != data->particles.size();i++ ) {
		Particle *p = data->particles[ i ];
		ParticleInfo pi = ParticleInfo( *p );
		CellId powner = p->getOwner();
		CellBorder *cb;
		cb = getBorder( powner, 0 );
		assert( !cb || pi.x() >= cb->getPos() );
		cb = getBorder( powner, 1 );
		assert( !cb || pi.x() <= cb->getPos() );
		cb = getBorder( powner, 2 );
		assert( !cb || pi.y() >= cb->getPos() );
		cb = getBorder( powner, 3 );
		assert( !cb || pi.y() <= cb->getPos() );
	};
};

void Simulation::synchronize( CellId ci ) {
	data->cells[ ci ].advanceTo( data->getTotalTime() );
};

bool Simulation::testOverlapping() {
	for ( int i = 0;i != data->particles.size();i++ ) {
		Particle *p = data->particles[ i ];
		CellId ci = p->getOwner();
		ParticleInfo p1( *p );
		for ( int i = 0;i < getNeighbourCount( ci );i++ ) {
			Cell& c = getCell( getNeighbour( ci, i ) );
			for ( List<Particle>::iterator it = c.getParticleIter();
			        it.valid();++it ) {
				ParticleInfo p2 = *it;
				if ( p != &( *it ) ) {
					double d = hypot( p1.x() - p2.x(), p1.y() - p2.y() );
					if ( d < p1.radius() + p2.radius() )
						return true;

				};
			};
		};
	};
	return false;
};

SimulationInfo::SimulationInfo( Simulation* sim0 ) : sim( sim0 ) {
	//first stop the cycle
	sim->data->should_stop_mutex.lock();
	sim->data->should_stop = true;
	//then wait for finish of run()
	sim->data->running.lock();
	sim->data->should_stop_mutex.unlock();
};

SimulationInfo::~SimulationInfo() {
	sim->data->running.unlock();
};

double SimulationInfo::getTime() {
	return sim->data->getTotalTime();
};

void Simulation::cellsInRectangle( vector<CellId>* v, const Rectangle& r ) const {
	double x1, y1, x2, y2;
	//we enlarge the rectangle in order to include neighbouring cells
	x1 = r.x1 - data->cellWidth;
	y1 = r.y1 - data->cellHeight;
	x2 = r.x2 + data->cellWidth;
	y2 = r.y2 + data->cellHeight;

	if ( x1 < data->minx )
		x1 = data->minx;
	if ( y1 < data->miny )
		y1 = data->miny;
	if ( x2 > data->maxx )
		x2 = data->maxx-eps;
	if ( y2 > data->maxy )
		y2 = data->maxy-eps;

	assert(x1<=x2); assert(y1<=y2);
	CellId left = findOwnerOf( x1, y1 );
	CellId right = findOwnerOf( x2, y1 );
	CellId bottom_left = findOwnerOf( x1, y2 );
	for ( ;left <= bottom_left;left += data->sqrtCellCount, right += data->sqrtCellCount )
		for ( CellId c = left;c <= right;c++ )
			v->push_back( c );
};

void Simulation::velocityChanged( const Object& o ) {
	/*	for each cell c where o is:
	for each particle p in c
	p.considerCollisionWith(o)
	/**/
	/*	vector<CellId> owners;
	owners.reserve(9);
	o.getOwners(*this,&owners);
	for (int i=0;i<owners.size();i++) {
	synchronize(owners[i]);
	Cell &owner=getCell(owners[i]);
	for (List<Particle>::iterator it=owner.getParticleIter();
	it.valid();++it) {
	it->considerCollisionWith(o);
		};
		};
	*/
}
;

double Simulation::getGravity() const
{
	return data->gravity;
}

ParticleProxy Simulation::getParticleProxy(int index)
{
	return ParticleProxy(*data->particles[index]);
}
