#ifndef SIMULATION_H
#define SIMULATION_H

#include "object.h"
#include "particle.h"
#include "wall.h"
#include "common.h"
#include "cell.h"
#include "myassert.h"
#include <vector>
#include <boost/intrusive_ptr.hpp>

struct SimulationData;
class Particle;
class ParticleInfo;
class Wall;
class Simulation;
class Cell;
class CellBorder;
class CollisionEvent;
class Event;
class Composite;

double computeCollisionTime(Particle &p1,Particle &p2);
double computeCollisionTime(Particle &p,Wall &w);
void doCollision(Particle &p1,Particle& p2);
void doCollision(Particle &p,Wall& w);

class SimulationError: public std::exception {}
;

class SimulationInfo;

class Simulation {
    friend class SimulationInfo;
    friend class Particle;
    friend class CollisionEvent;
private:
    bool initialized;
    void initialize();
    SimulationData *data;
	Simulation(const Simulation&); // non-copyable 

    void writeLog();
    void run1();
    double getNextCollisionTime();
    //following functions should be called via SimulationInfo
    double getMaxX();
    double getMaxY();
public:
	void invalidateCollisionersOf(Object& ,CellId);
    ParticleInfo getParticle(int index);
    Wall& getWall(int index);
    Object& getComposite(int index);

    Simulation(int particleCount,int wallCount,int CellCount,double maxx,double maxy,double gravity);
    ~Simulation();
    //runs the simulation, advancing the simulation time by dt
    void advanceBy(double dt);
    //runs the simulation until stop() is called
    void run();
    //stops running simulation; doesn't block until simulation actually stops
    void stop();
    //the simulation takes ownership of those objects passed via addObject
    void addObject(Particle* p);
    void addObject(Wall* w);
    void addObject(Composite* o);
    int getParticleCount();
    int getWallCount();
    int getCompositeCount();
    double getTotalTime();
    int getCollisionCount();
    // finds the cell owning the point (x,y)
    CellId findOwnerOf(double x,double y) const;
    // finds the cell owning p
    CellId findOwnerOf(Particle const &p) const;
    int getNeighbourCount(CellId c) const {
        return 9;
    };
    CellId getNeighbour(CellId ci,int index) const;
    Cell& getCell(CellId ci);
    bool valid(CellId ci);
    int getBorderCount(CellId ci) {
        return 4;
    }
    CellBorder* getBorder(CellId ci,int index);
    void synchronize(CellId ci);
    bool test();
    bool testOverlapping();

    void addEvent(boost::intrusive_ptr<Event> e);
    void removeEvent(boost::intrusive_ptr<Event> e);
    // adds to v(with v->push_back(...)) all cells in rectangle (x1,y1,x2,y2)
    // including neighbours of cells in that rectangle
    // v must be a initialized vector
    void cellsInRectangle(std::vector<CellId>* v,const Rectangle& r) const;
    // velocityChanged is called after an object  o changed its velocity
    void velocityChanged(const Object& o);
    double getGravity() const;
    ParticleProxy getParticleProxy(int index);
	
	
};

class SimulationInfo {
public:
    SimulationInfo(Simulation* sim0);
    ~SimulationInfo();
    double getMaxX() {
        return sim->getMaxX();
    }
    double getMaxY() {
        return sim->getMaxY();
    }
    ParticleInfo getParticle(int index) {
        return sim->getParticle(index);
    }
	ParticleProxy getParticleProxy(int index) {
		return sim->getParticleProxy(index);
	};
    WallInfo getWall(int index) {
        return sim->getWall(index);
    }
    Object& getComposite(int index) {
        return sim->getComposite(index);
    }
    double getTime();
private:
    Simulation *sim;
};

#endif // SIMULATION_H
