#include "simulation.h"
#include "gui/canvas.h"
#include <vector>
#include <limits>

Wall::~Wall() {}

void Wall::writeLog( FILE* file) {
	fprintf( file, "id: %2d, pos: %lf, type: %s\n",
	         getId(), pos, ( type == wHorizontal ) ? "horizontal" : "vertical" );
};

void Wall::collisionDirection( const Particle& p, double &dx, double &dy ) {
	if ( type == wHorizontal ) {
		dx = 0;
		dy = 1;
	} else {
		dx = 1;
		dy = 0;
	}
};

bool Wall::testCollision( Particle const  &p ) const {
	ParticleInfo pi( p );
	double px = ( type == wHorizontal ) ? pi.y() : pi.x();
	double py = ( type == wHorizontal ) ? pi.x() : pi.y();
	return equal( fabs( px -pos ), pi.radius() ) && start < py && py < end;
};

int convert(double x) {
	if (x==infinity) return 10000;//std::numeric_limits<int>::max();
	if (x==-infinity) return -10000;//std::numeric_limits<int>::min();
	return int(x);
};

void Wall::draw( Canvas& c ) const {
	if ( type == wHorizontal )
		c.drawLine( convert( start ), int( pos ), convert( end ), int( pos ) );
	else
		c.drawLine( int( pos ), convert( start ), int( pos ), convert( end ) );
};

Rectangle Wall::boundingRect() const {
	if ( type == wHorizontal )
		return Rectangle( start, pos, end, pos );
	else
		return Rectangle( pos, start, pos, end );
};

void Wall::getOwners(std::vector<CellId> *v ) const {
	Rectangle r = boundingRect();
	sim->cellsInRectangle( v, r );
};

void Wall::run(double dt) {
	// if it ignores gravity, then it runs as if gravity==0
	double gravity=ignoreGravity ? 0 : sim->getGravity();
	double gravdx=gravity*dt*dt/2;
	if (type==wHorizontal) {
		start+=vx*dt; end+=vx*dt;
		pos+=vy*dt-gravdx;
	} else {
		start+=vy*dt-gravdx; end+=vy*dt-gravdx;
		pos+=vx*dt;
	};
	vy-=gravity*dt;
};

Wall::Wall( Simulation * sim0, WallType type0, double pos0, double start0, double end0, double m0, bool ignoreGravity0 ):Object(sim0),pos(pos0),type(type0),m(m0),
		start(start0),end(end0),temperature(0),heatTransfer(0),ignoreGravity(ignoreGravity0)
{
	// so that we can assume that start<end
	if (end<start) {
		double temp=end;
		end=start;
		start=temp;
	};
};

template<typename T> inline void sort2(T& a, T& b) {
	if (a>b) {
		T p=a;
		a=b;
		b=p;
	};
};



/* The idea is first to compute the collision time as if the wall was unlimited,
and then check the collision point, whether it is in limit.
colTimeHelper does the first part. 
*/
// wpos - position of wall
// px - position of particle
// pv  - velocity of particle
// pr - radius of particle
double colTimeHelper(double wpos,double px,double pv,double pr,double g) {
	double A=-g/2;
	double B=pv;
	double C;
	if (A==0) {
		if (pv==0) return infinity;
		double t=(wpos-px)/pv;    //t je "stredny" cas, t=(t1+t2)/2
		if (t>0) {
			double t1=(wpos-px+pr)/pv;
			double t2=(wpos-px-pr)/pv;
			return t1<t2 ? t1 : t2;
		} else
			return infinity;
	} else {
		double D,t1,t2,t3,t4;
		t1=t2=t3=t4=infinity;

		C=px-wpos-pr;
		D=B*B-4*A*C;
		if (D>=0) {
			t1=(-B-sqrt(D))/2/A;
			t4=(-B+sqrt(D))/2/A;
		};

		C=px-wpos+pr;
		D=B*B-4*A*C;
		if (D>=0) {
			t2=(-B-sqrt(D))/2/A;
			t3=(-B+sqrt(D))/2/A;
		};
		sort2(t1,t2);
		sort2(t2,t3);
		sort2(t3,t4);
		sort2(t1,t2);
		sort2(t2,t3);
		sort2(t1,t2);
		assert(t1<=t2);assert(t2<=t3);assert(t3<=t4);

		// which collision time is the closest to 0?
		if (t1+t2>0) return t1;  // is current time before the middle of <t1,t2> => return t1
		if (t3+t4>0) return t3;  // before the middle of <t3,t4> => return t3
		return infinity;         // after => return infinity
		
/*
		double t=t1;
		if (fabs(t2)<fabs(t)) t=t2;
		if (fabs(t3)<fabs(t)) t=t3;
		if (fabs(t4)<fabs(t)) t=t4;

		if (t==t1) return t1;
		if (t==t2) return t3;
		if (t==t3) return t3; 
		if (t==t4) return infinity;
	
		assert(false); //shouldn`t happen
		return 0; //only to supress warning
*/
	} 
	
}
double computeCollisionTime(const Particle &p,const Wall &w) {
	// we reduce both cases to case where x axis is perpendicular to the wall
	double px,pvx,gx,py,pvy,gy;
	// if w ignores gravity, the we compute with gravity=p.sim.getGravity(), because we are interested in the difference(p undergoes gravity and w does not). If w does not ignore gravity. p's movement relative to w is uniform and straight. (Particles never ignore gravitation)
	double gravity=w.ignoreGravity ? p.sim->getGravity() : 0;

	int dcount=0;
	int dval=w.type;
	if (w.type==wVertical) {
		px=p.x; pvx=p.vx-w.vx; gx=0; py=p.y; pvy=p.vy-w.vy; gy=gravity;
		dcount++;
	};
	if (w.type==wHorizontal) {
		py=p.x; pvy=p.vx-w.vx; gy=0; px=p.y; pvx=p.vy-w.vy; gx=gravity;
		dcount++;
	};
	assert(dcount>0);
	// no collision occurs if the particle is already overlapping with they
	// extended wall but not with the wall.
	if (!(w.start<py && py<w.end))  // out of y limits
		if (w.pos-p.radius()<px && px<w.pos+p.radius()) //but would overlap if the wall was infinite
			return infinity;
	double t=colTimeHelper(w.pos,px,pvx,p.radius(),gx);
	//the position of the particle on y axis(ie. parallel with the wall) 
	// when it hits the wall
	double py1=py+pvy*t;
	if (w.start<py1 && py1<w.end)
		return t;   //collision in (start,end) => valid
	else
		return infinity;  // invalid
	
};

void Wall::setTemperature( double temp )
{
	temperature=temp;
}

void Wall::setHeatTransfer( double ht )
{
	assert(0<=ht && ht<=1 );
	heatTransfer=ht;
}

// overriden to implement heat transfers
void Wall::doCollisionPrivate( Particle & p )
{
	Object::doCollisionPrivate(p);
	double v0=hypot(p.vx,p.vy);
	double v=(v0*(1-heatTransfer) + temperature*heatTransfer);
	double c=v/v0;
	p.vx*=c;
	p.vy*=c;
}
